import tempfile
import urllib.request
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Union

from ops.model import _ModelBackend
from ops.pebble import Client, ExecError

from scenario.logger import logger as scenario_logger

if TYPE_CHECKING:
    from scenario.state import Event, ExecOutput, State, _CharmSpec

logger = scenario_logger.getChild("mocking")


class _MockExecProcess:
    def __init__(self, command: Tuple[str], change_id: int, out: "ExecOutput"):
        self._command = command
        self._change_id = change_id
        self._out = out
        self._waited = False
        self.stdout = StringIO(self._out.stdout)
        self.stderr = StringIO(self._out.stderr)

    def wait(self):
        self._waited = True
        exit_code = self._out.return_code
        if exit_code != 0:
            raise ExecError(list(self._command), exit_code, None, None)

    def wait_output(self):
        out = self._out
        exit_code = out.return_code
        if exit_code != 0:
            raise ExecError(list(self._command), exit_code, None, None)
        return out.stdout, out.stderr

    def send_signal(self, sig: Union[int, str]):
        pass


class _MockModelBackend(_ModelBackend):
    def __init__(self, state: "State", event: "Event", charm_spec: "_CharmSpec"):
        super().__init__(state.unit_name, state.model.name, state.model.uuid)
        self._state = state
        self._event = event
        self._charm_spec = charm_spec

    def get_pebble(self, socket_path: str) -> "Client":
        return _MockPebbleClient(
            socket_path=socket_path,
            state=self._state,
            event=self._event,
            charm_spec=self._charm_spec,
        )

    def _get_relation_by_id(self, rel_id):
        try:
            return next(
                filter(lambda r: r.relation_id == rel_id, self._state.relations)
            )
        except StopIteration as e:
            raise RuntimeError(f"Not found: relation with id={rel_id}.") from e

    def relation_get(self, rel_id, obj_name, app):
        relation = self._get_relation_by_id(rel_id)
        if app and obj_name == self._state.app_name:
            return relation.local_app_data
        elif app:
            return relation.remote_app_data
        elif obj_name == self._state.unit_name:
            return relation.local_unit_data
        else:
            unit_id = obj_name.split("/")[-1]
            return relation.remote_units_data[int(unit_id)]

    def is_leader(self):
        return self._state.leader

    def status_get(self, *args, **kwargs):
        status, message = (
            self._state.status.app if kwargs.get("app") else self._state.status.unit
        )
        return {"status": status, "message": message}

    def relation_ids(self, endpoint, *args, **kwargs):
        return [
            rel.relation_id for rel in self._state.relations if rel.endpoint == endpoint
        ]

    def relation_list(self, rel_id, *args, **kwargs):
        relation = self._get_relation_by_id(rel_id)
        return tuple(
            f"{relation.remote_app_name}/{unit_id}"
            for unit_id in relation.remote_unit_ids
        )

    def config_get(self, *args, **kwargs):
        state_config = self._state.config
        if not state_config:
            state_config = {
                key: value.get("default")
                for key, value in self._charm_spec.config.items()
            }

        if args:  # one specific key requested
            # Fixme: may raise KeyError if the key isn't defaulted. What do we do then?
            return state_config[args[0]]

        return state_config  # full config

    def network_get(self, *args, **kwargs):
        name, relation_id = args

        network = next(filter(lambda r: r.name == name, self._state.networks))
        return network.hook_tool_output_fmt()

    def action_get(self, *args, **kwargs):
        raise NotImplementedError("action_get")

    def relation_remote_app_name(self, *args, **kwargs):
        raise NotImplementedError("relation_remote_app_name")

    def resource_get(self, *args, **kwargs):
        raise NotImplementedError("resource_get")

    def storage_list(self, *args, **kwargs):
        raise NotImplementedError("storage_list")

    def storage_get(self, *args, **kwargs):
        raise NotImplementedError("storage_get")

    def planned_units(self, *args, **kwargs):
        raise NotImplementedError("planned_units")

    # setter methods: these can mutate the state.
    def application_version_set(self, *args, **kwargs):
        self._state.status.app_version = args[0]
        return None

    def status_set(self, *args, **kwargs):
        if kwargs.get("is_app"):
            self._state.status.app = args
        else:
            self._state.status.unit = args
        return None

    def juju_log(self, *args, **kwargs):
        self._state.juju_log.append(args)
        return None

    def relation_set(self, *args, **kwargs):
        rel_id, key, value, app = args
        relation = self._get_relation_by_id(rel_id)
        if app:
            if not self._state.leader:
                raise RuntimeError("needs leadership to set app data")
            tgt = relation.local_app_data
        else:
            tgt = relation.local_unit_data
        tgt[key] = value
        return None

    # TODO:
    def action_set(self, *args, **kwargs):
        raise NotImplementedError("action_set")

    def action_fail(self, *args, **kwargs):
        raise NotImplementedError("action_fail")

    def action_log(self, *args, **kwargs):
        raise NotImplementedError("action_log")

    def storage_add(self, *args, **kwargs):
        raise NotImplementedError("storage_add")

    def secret_get(
        self,
        *,
        id: str = None,
        label: str = None,
        refresh: bool = False,
        peek: bool = False,
    ) -> Dict[str, str]:
        # cleanup id:
        if id and id.startswith("secret:"):
            id = id[7:]

        if id:
            try:
                secret = next(filter(lambda s: s.id == id, self._state.secrets))
            except StopIteration:
                raise RuntimeError(f"not found: secret with id={id}.")
        elif label:
            try:
                secret = next(filter(lambda s: s.label == label, self._state.secrets))
            except StopIteration:
                raise RuntimeError(f"not found: secret with label={label}.")
        else:
            raise RuntimeError(f"need id or label.")

        revision = secret.revision
        if peek or refresh:
            revision = max(secret.contents.keys())
            if refresh:
                secret.revision = revision

        return secret.contents[revision]

    def secret_set(self, *args, **kwargs):
        raise NotImplementedError("secret_set")

    def secret_grant(self, *args, **kwargs):
        raise NotImplementedError("secret_grant")

    def secret_remove(self, *args, **kwargs):
        raise NotImplementedError("secret_remove")


class _MockPebbleClient(Client):
    def __init__(
        self,
        socket_path: str,
        opener: Optional[urllib.request.OpenerDirector] = None,
        base_url: str = "http://localhost",
        timeout: float = 5.0,
        *,
        state: "State",
        event: "Event",
        charm_spec: "_CharmSpec",
    ):
        super().__init__(socket_path, opener, base_url, timeout)
        self._state = state
        self._event = event
        self._charm_spec = charm_spec

    @property
    def _container(self):
        container_name = self.socket_path.split("/")[-2]
        try:
            return next(
                filter(lambda x: x.name == container_name, self._state.containers)
            )
        except StopIteration:
            raise RuntimeError(
                f"container with name={container_name!r} not found. "
                f"Did you forget a Container, or is the socket path "
                f"{self.socket_path!r} wrong?"
            )

    def _request(self, *args, **kwargs):
        if args == ("GET", "/v1/system-info"):
            if self._container.can_connect:
                return {"result": {"version": "unknown"}}
            else:
                wrap_errors = False  # this is what Client expects!
                raise FileNotFoundError("")

        elif args[:2] == ("GET", "/v1/services"):
            service_names = list(args[2]["names"].split(","))
            result = []

            for layer in self._container.layers:
                if not service_names:
                    break

                for name in service_names:
                    if name in layer["services"]:
                        service_names.remove(name)
                        result.append(layer["services"][name])

            # todo: what do we do if we don't find the requested service(s)?
            return {"result": result}

        else:
            raise NotImplementedError(f"_request: {args}")

    def exec(self, *args, **kwargs):
        cmd = tuple(args[0])
        out = self._container.exec_mock.get(cmd)
        if not out:
            raise RuntimeError(f"mock for cmd {cmd} not found.")

        change_id = out._run()
        return _MockExecProcess(change_id=change_id, command=cmd, out=out)

    def pull(self, *args, **kwargs):
        # todo double-check how to surface error
        wrap_errors = False

        path_txt = args[0]
        pos = self._container.filesystem
        for token in path_txt.split("/")[1:]:
            pos = pos.get(token)
            if not pos:
                raise FileNotFoundError(path_txt)
        local_path = Path(pos)
        if not local_path.exists() or not local_path.is_file():
            raise FileNotFoundError(local_path)
        return local_path.open()

    def push(self, *args, **kwargs):
        setter = True
        # todo double-check how to surface error
        wrap_errors = False

        path_txt, contents = args

        pos = self._container.filesystem
        tokens = path_txt.split("/")[1:]
        for token in tokens[:-1]:
            nxt = pos.get(token)
            if not nxt and kwargs["make_dirs"]:
                pos[token] = {}
                pos = pos[token]
            elif not nxt:
                raise FileNotFoundError(path_txt)
            else:
                pos = pos[token]

        # dump contents
        # fixme: memory leak here if tmp isn't regularly cleaned up
        file = tempfile.NamedTemporaryFile(delete=False)
        pth = Path(file.name)
        pth.write_text(contents)

        pos[tokens[-1]] = pth
        return



import copy
import dataclasses
import inspect
import typing
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
from uuid import uuid4

import yaml

from scenario.logger import logger as scenario_logger
from scenario.runtime import trigger

if typing.TYPE_CHECKING:
    try:
        from typing import Self
    except ImportError:
        from typing_extensions import Self
    from ops.pebble import LayerDict
    from ops.testing import CharmType

logger = scenario_logger.getChild("structs")

ATTACH_ALL_STORAGES = "ATTACH_ALL_STORAGES"
CREATE_ALL_RELATIONS = "CREATE_ALL_RELATIONS"
BREAK_ALL_RELATIONS = "BREAK_ALL_RELATIONS"
DETACH_ALL_STORAGES = "DETACH_ALL_STORAGES"
META_EVENTS = {
    "CREATE_ALL_RELATIONS": "-relation-created",
    "BREAK_ALL_RELATIONS": "-relation-broken",
    "DETACH_ALL_STORAGES": "-storage-detaching",
    "ATTACH_ALL_STORAGES": "-storage-attached",
}


@dataclasses.dataclass
class _DCBase:
    def replace(self, *args, **kwargs):
        return dataclasses.replace(self, *args, **kwargs)

    def copy(self) -> "Self":
        return copy.deepcopy(self)


@dataclasses.dataclass
class Secret(_DCBase):
    id: str

    # mapping from revision IDs to each revision's contents
    contents: Dict[int, Dict[str, str]]

    owned_by_this_unit: bool = False

    # has this secret been granted to this unit/app or neither?
    granted: Literal["unit", "app", False] = False

    # what revision is currently tracked by this charm. Only meaningful if owned_by_this_unit=False
    revision: int = 0

    label: str = None

    # consumer-only events
    @property
    def changed_event(self):
        """Sugar to generate a secret-changed event."""
        if self.owned_by_this_unit:
            raise ValueError(
                "This unit will never receive secret-changed for a secret it owns."
            )
        return Event(name="secret-changed", secret=self)

    # owner-only events
    @property
    def rotate_event(self):
        """Sugar to generate a secret-rotate event."""
        if not self.owned_by_this_unit:
            raise ValueError(
                "This unit will never receive secret-rotate for a secret it does not own."
            )
        return Event(name="secret-rotate", secret=self)

    @property
    def expired_event(self):
        """Sugar to generate a secret-expired event."""
        if not self.owned_by_this_unit:
            raise ValueError(
                "This unit will never receive secret-expire for a secret it does not own."
            )
        return Event(name="secret-expire", secret=self)

    @property
    def remove_event(self):
        """Sugar to generate a secret-remove event."""
        if not self.owned_by_this_unit:
            raise ValueError(
                "This unit will never receive secret-removed for a secret it does not own."
            )
        return Event(name="secret-removed", secret=self)


_RELATION_IDS_CTR = 0


@dataclasses.dataclass
class Relation(_DCBase):
    endpoint: str
    remote_app_name: str
    remote_unit_ids: List[int] = dataclasses.field(default_factory=list)

    # local limit
    limit: int = 1

    # scale of the remote application; number of units, leader ID?
    # TODO figure out if this is relevant
    scale: int = 1
    leader_id: int = 0

    # we can derive this from the charm's metadata
    interface: str = None

    # Every new Relation instance gets a new one, if there's trouble, override.
    relation_id: int = -1

    local_app_data: Dict[str, str] = dataclasses.field(default_factory=dict)
    remote_app_data: Dict[str, str] = dataclasses.field(default_factory=dict)
    local_unit_data: Dict[str, str] = dataclasses.field(default_factory=dict)
    remote_units_data: Dict[int, Dict[str, str]] = dataclasses.field(
        default_factory=dict
    )

    def __post_init__(self):
        global _RELATION_IDS_CTR
        if self.relation_id == -1:
            _RELATION_IDS_CTR += 1
            self.relation_id = _RELATION_IDS_CTR

        if self.remote_unit_ids and self.remote_units_data:
            if not set(self.remote_unit_ids) == set(self.remote_units_data):
                raise ValueError(
                    f"{self.remote_unit_ids} should include any and all IDs from {self.remote_units_data}"
                )
        elif self.remote_unit_ids:
            self.remote_units_data = {x: {} for x in self.remote_unit_ids}
        elif self.remote_units_data:
            self.remote_unit_ids = [x for x in self.remote_units_data]
        else:
            self.remote_unit_ids = [0]
            self.remote_units_data = {0: {}}

    @property
    def changed_event(self):
        """Sugar to generate a <this relation>-relation-changed event."""
        return Event(name=self.endpoint + "-relation-changed", relation=self)

    @property
    def joined_event(self):
        """Sugar to generate a <this relation>-relation-joined event."""
        return Event(name=self.endpoint + "-relation-joined", relation=self)

    @property
    def created_event(self):
        """Sugar to generate a <this relation>-relation-created event."""
        return Event(name=self.endpoint + "-relation-created", relation=self)

    @property
    def departed_event(self):
        """Sugar to generate a <this relation>-relation-departed event."""
        return Event(name=self.endpoint + "-relation-departed", relation=self)

    @property
    def broken_event(self):
        """Sugar to generate a <this relation>-relation-broken event."""
        return Event(name=self.endpoint + "-relation-broken", relation=self)


def _random_model_name():
    import random
    import string

    space = string.ascii_letters + string.digits
    return "".join(random.choice(space) for _ in range(20))


@dataclasses.dataclass
class Model(_DCBase):
    name: str = _random_model_name()
    uuid: str = str(uuid4())


_SimpleFS = Dict[
    str,  # file/dirname
    Union["_SimpleFS", Path],  # subdir  # local-filesystem path resolving to a file.
]

# for now, proc mock allows you to map one command to one mocked output.
# todo extend: one input -> multiple outputs, at different times


_CHANGE_IDS = 0


@dataclasses.dataclass
class ExecOutput:
    return_code: int = 0
    stdout: str = ""
    stderr: str = ""

    # change ID: used internally to keep track of mocked processes
    _change_id: int = -1

    def _run(self) -> int:
        global _CHANGE_IDS
        _CHANGE_IDS = self._change_id = _CHANGE_IDS + 1
        return _CHANGE_IDS


_ExecMock = Dict[Tuple[str, ...], ExecOutput]


@dataclasses.dataclass
class Container(_DCBase):
    name: str
    can_connect: bool = False
    layers: Tuple["LayerDict"] = ()

    # this is how you specify the contents of the filesystem: suppose you want to express that your
    # container has:
    # - /home/foo/bar.py
    # - /bin/bash
    # - /bin/baz
    #
    # this becomes:
    # filesystem = {
    #     'home': {
    #         'foo': Path('/path/to/local/file/containing/bar.py')
    #     },
    #     'bin': {
    #         'bash': Path('/path/to/local/bash'),
    #         'baz': Path('/path/to/local/baz')
    #     }
    # }
    # when the charm runs `pebble.pull`, it will return .open() from one of those paths.
    # when the charm pushes, it will either overwrite one of those paths (careful!) or it will
    # create a tempfile and insert its path in the mock filesystem tree
    # charm-created tempfiles will NOT be automatically deleted -- you have to clean them up yourself!
    filesystem: _SimpleFS = dataclasses.field(default_factory=dict)

    exec_mock: _ExecMock = dataclasses.field(default_factory=dict)

    @property
    def pebble_ready_event(self):
        """Sugar to generate a <this container's name>-pebble-ready event."""
        if not self.can_connect:
            logger.warning(
                "you **can** fire pebble-ready while the container cannot connect, "
                "but that's most likely not what you want."
            )
        return Event(name=self.name + "-pebble-ready", container=self)


@dataclasses.dataclass
class Address(_DCBase):
    hostname: str
    value: str
    cidr: str


@dataclasses.dataclass
class BindAddress(_DCBase):
    mac_address: str
    interface_name: str
    interfacename: str  # noqa legacy
    addresses: List[Address]

    def hook_tool_output_fmt(self):
        # dumps itself to dict in the same format the hook tool would
        return {
            "bind-addresses": self.mac_address,
            "interface-name": self.interface_name,
            "interfacename": self.interfacename,
            "addresses": [dataclasses.asdict(addr) for addr in self.addresses],
        }


@dataclasses.dataclass
class Network(_DCBase):
    name: str
    bind_id: int

    bind_addresses: List[BindAddress]
    bind_address: str
    egress_subnets: List[str]
    ingress_addresses: List[str]

    is_default: bool = False

    def hook_tool_output_fmt(self):
        # dumps itself to dict in the same format the hook tool would
        return {
            "bind-addresses": [ba.hook_tool_output_fmt() for ba in self.bind_addresses],
            "bind-address": self.bind_address,
            "egress-subnets": self.egress_subnets,
            "ingress-addresses": self.ingress_addresses,
        }

    @classmethod
    def default(
        cls,
        name,
        bind_id,
        private_address: str = "1.1.1.1",
        mac_address: str = "",
        hostname: str = "",
        cidr: str = "",
        interface_name: str = "",
        egress_subnets=("1.1.1.2/32",),
        ingress_addresses=("1.1.1.2",),
    ) -> "Network":
        """Helper to create a minimal, heavily defaulted Network."""
        return cls(
            name=name,
            bind_id=bind_id,
            bind_addresses=[
                BindAddress(
                    mac_address=mac_address,
                    interface_name=interface_name,
                    interfacename=interface_name,
                    addresses=[
                        Address(hostname=hostname, value=private_address, cidr=cidr)
                    ],
                )
            ],
            bind_address=private_address,
            egress_subnets=list(egress_subnets),
            ingress_addresses=list(ingress_addresses),
        )


@dataclasses.dataclass
class Status(_DCBase):
    app: Tuple[str, str] = ("unknown", "")
    unit: Tuple[str, str] = ("unknown", "")
    app_version: str = ""


@dataclasses.dataclass
class State(_DCBase):
    config: Dict[str, Union[str, int, float, bool]] = None
    relations: Sequence[Relation] = dataclasses.field(default_factory=list)
    networks: Sequence[Network] = dataclasses.field(default_factory=list)
    containers: Sequence[Container] = dataclasses.field(default_factory=list)
    status: Status = dataclasses.field(default_factory=Status)
    leader: bool = False
    model: Model = Model()
    juju_log: Sequence[Tuple[str, str]] = dataclasses.field(default_factory=list)
    secrets: List[Secret] = dataclasses.field(default_factory=list)

    # meta stuff: actually belongs in event data structure.
    juju_version: str = "3.0.0"
    unit_id: str = "0"
    app_name: str = "local"

    # todo:
    #  actions?

    @property
    def unit_name(self):
        return self.app_name + "/" + self.unit_id

    def with_can_connect(self, container_name: str, can_connect: bool):
        def replacer(container: Container):
            if container.name == container_name:
                return container.replace(can_connect=can_connect)
            return container

        ctrs = tuple(map(replacer, self.containers))
        return self.replace(containers=ctrs)

    def with_leadership(self, leader: bool):
        return self.replace(leader=leader)

    def with_unit_status(self, status: str, message: str):
        return self.replace(
            status=dataclasses.replace(self.status, unit=(status, message))
        )

    def get_container(self, name) -> Container:
        try:
            return next(filter(lambda c: c.name == name, self.containers))
        except StopIteration as e:
            raise ValueError(f"container: {name}") from e

    # FIXME: not a great way to obtain a delta, but is "complete" todo figure out a better way.
    def jsonpatch_delta(self, other: "State"):
        try:
            import jsonpatch
        except ModuleNotFoundError:
            logger.error(
                "cannot import jsonpatch: using the .delta() "
                "extension requires jsonpatch to be installed."
                "Fetch it with pip install jsonpatch."
            )
            return NotImplemented
        patch = jsonpatch.make_patch(
            dataclasses.asdict(other), dataclasses.asdict(self)
        ).patch
        return sort_patch(patch)

    def trigger(
        self,
        event: Union["Event", str],
        charm_type: Type["CharmType"],
        # callbacks
        pre_event: Optional[Callable[["CharmType"], None]] = None,
        post_event: Optional[Callable[["CharmType"], None]] = None,
        # if not provided, will be autoloaded from charm_type.
        meta: Optional[Dict[str, Any]] = None,
        actions: Optional[Dict[str, Any]] = None,
        config: Optional[Dict[str, Any]] = None,
    ):
        """Fluent API for trigger."""
        return trigger(
            state=self,
            event=event,
            charm_type=charm_type,
            pre_event=pre_event,
            post_event=post_event,
            meta=meta,
            actions=actions,
            config=config,
        )


@dataclasses.dataclass
class _CharmSpec(_DCBase):
    """Charm spec."""

    charm_type: Type["CharmType"]
    meta: Optional[Dict[str, Any]]
    actions: Optional[Dict[str, Any]] = None
    config: Optional[Dict[str, Any]] = None

    @staticmethod
    def autoload(charm_type: Type["CharmType"]):
        charm_source_path = Path(inspect.getfile(charm_type))
        charm_root = charm_source_path.parent.parent

        metadata_path = charm_root / "metadata.yaml"
        meta = yaml.safe_load(metadata_path.open())

        actions = config = None

        config_path = charm_root / "config.yaml"
        if config_path.exists():
            config = yaml.safe_load(config_path.open())

        actions_path = charm_root / "actions.yaml"
        if actions_path.exists():
            actions = yaml.safe_load(actions_path.open())

        return _CharmSpec(
            charm_type=charm_type, meta=meta, actions=actions, config=config
        )


def sort_patch(patch: List[Dict], key=lambda obj: obj["path"] + obj["op"]):
    return sorted(patch, key=key)


@dataclasses.dataclass
class Event(_DCBase):
    name: str
    args: Tuple[Any] = ()
    kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)

    # if this is a relation event, the relation it refers to
    relation: Optional[Relation] = None

    # if this is a secret event, the secret it refers to
    secret: Optional[Secret] = None

    # if this is a workload (container) event, the container it refers to
    container: Optional[Container] = None

    # todo add other meta for
    #  - secret events
    #  - pebble?
    #  - action?


@dataclasses.dataclass
class Inject(_DCBase):
    """Base class for injectors: special placeholders used to tell harness_ctx
    to inject instances that can't be retrieved in advance in event args or kwargs.
    """

    pass


@dataclasses.dataclass
class InjectRelation(Inject):
    relation_name: str
    relation_id: Optional[int] = None


def _derive_args(event_name: str):
    args = []
    terms = {
        "-relation-changed",
        "-relation-broken",
        "-relation-joined",
        "-relation-departed",
        "-relation-created",
    }

    for term in terms:
        # fixme: we can't disambiguate between relation IDs.
        if event_name.endswith(term):
            args.append(InjectRelation(relation_name=event_name[: -len(term)]))

    return tuple(args)
